package main

import (
	"container/heap"
	"encoding/json"
	"errors"
	"fmt"
	"github.com/jasonlvhit/gocron"
	"io"
	"log"
	"os"
	"regexp"
	"strconv"
	"strings"
	"time"
)

var songs SongPriorityQueue

var credentials struct {
	APIURL   string
	UserName string
	Password string
}

func appendSong(wikiText string, author string, newSong string) string {
	const AUTHOR_MARK = "<!-- Lisääjä -->"
	const SONG_MARK = "<!-- Kappale -->"
	lines := strings.Split(wikiText, "\n")
	authorPrevIndex := -2

	changedLines := make([]string, 0, len(lines))
	for index, line := range lines {
		if strings.Index(line, AUTHOR_MARK) != -1 && strings.Index(line, author) != -1 {
			authorPrevIndex = index
		}
		if authorPrevIndex == (index-1) && strings.Index(line, SONG_MARK) != -1 {
			changedLines = append(changedLines, "| "+SONG_MARK+" "+newSong)
		} else {
			changedLines = append(changedLines, line)
		}
	}
	return strings.Join(changedLines, "\n")
}

func addSong(title string, week int, author string, song string) (bool, error) {
	wiki := CreateWikiClient(credentials.APIURL, credentials.UserName, credentials.Password)

	sections, err := wiki.GetWikiPageSections(title)
	if err != nil {
		return false, err
	}

	numberReg, _ := regexp.Compile("\\d+")
	for _, section := range sections {
		weekStr := numberReg.FindString(section.title)
		if weekStr != "" {
			weekNumber, _ := strconv.Atoi(weekStr)
			if weekNumber == week {
				wikiText, err := wiki.GetWikiPageSectionText(title, section.index)
				if err != nil {
					return false, err
				}
				changedWikiText := appendSong(wikiText, author, song)

				return wiki.EditWikiPageSection(title, section.index, changedWikiText,
					fmt.Sprintf("Added week %d song for %s", week, author))
			}
		}
	}
	return false, errors.New("Could not find matching section")
}

func songSynced(syncedWeek int) error {
	v, err := loadDb()
	if err != nil {
		return err
	}

	synced := false
	for _, song := range v.Songs {
		if song.Week == syncedWeek {
			song.Sync = true
			synced = true
			err := saveDb(v)
			if err != nil {
				return err
			}
		}
	}
	if !synced {
		return errors.New("No week matched from JSON for synced song")
	}
	return errors.New("Week not found")
}

func getSongs() (SongPriorityQueue, error) {
	dbSongs, err := loadDb()
	if err != nil {
		return nil, err
	}

	songs := make(SongPriorityQueue, len(dbSongs.Songs))
	for index, songObj := range dbSongs.Songs {
		songs[index] = &Song{
			time:  targetTime(songObj),
			song:  songEntryWikiText(songObj),
			week:  songObj.Week,
			sync:  songObj.Sync,
			index: index,
		}
	}
	heap.Init(&songs)
	for len(songs) > 0 && songs[0].sync {
		heap.Pop(&songs)
	}
	return songs, nil
}

func task() {
	now := time.Now()
	if len(songs) > 0 && songs[0].time.Before(now) {
		fmt.Println("Time has passed for " + songs[0].song)
		success, err := addSong("Levyraati 2018", songs[0].week, "Lamperi", songs[0].song)
		if err != nil {
			log.Fatal(err)
		}
		if success {
			err := songSynced(songs[0].week)
			if err == nil {
				heap.Pop(&songs)
			} else {
				fmt.Println("Error received:", err)
			}
		}
	}
}

func initCreds() error {
	f, err := os.Open("credentials.json")
	if err != nil {
		return err
	}
	defer f.Close()

	if err != nil {
		log.Fatal(err)
		return err
	}
	dec := json.NewDecoder(f)
	for {
		if err := dec.Decode(&credentials); err == io.EOF {
			break
		} else if err != nil {
			log.Fatal(err)
		}
	}
	return nil
}

func targetTime(entry *SongEntry) time.Time {
	yearStart, _ := time.Parse(time.RFC3339, "2018-01-01T00:00:00+02:00")
	target := yearStart.AddDate(0, 0, (entry.Week-1)*7)
	return target
}

func songEntryWikiText(entry *SongEntry) string {
	return songWikiText(entry.URL, entry.Artist, entry.Title)
}

func songWikiText(url string, artist string, title string) string {
	return "[" + url + " " + artist + " - " + title + "]"
}

func main() {
	err := initCreds()
	if err != nil {
		panic(err)
	}
	songs, err = getSongs()
	if err != nil {
		panic(err)
	}

	modifiedSongChan := make(chan *SongEntry)

	go func() {
		webStart(modifiedSongChan)
	}()

	go func() {
		for {
			newSong := <-modifiedSongChan
			matched := false
			for _, song := range songs {
				if song.week == newSong.Week {
					song.song = songEntryWikiText(newSong)
					song.sync = newSong.Sync
					matched = true
					log.Printf("Updated song for week %d, artist: %s, title: %s, URL: %s, time: %v",
						newSong.Week, newSong.Artist, newSong.Title, newSong.URL, song.time)
				}
			}
			if !matched {
				song := &Song{
					time:  targetTime(newSong),
					song:  songEntryWikiText(newSong),
					week:  newSong.Week,
					sync:  newSong.Sync,
					index: len(songs),
				}
				heap.Push(&songs, song)
				log.Printf("Added song for week %d, artist: %s, title: %s, URL: %s, time: %v",
					newSong.Week, newSong.Artist, newSong.Title, newSong.URL, song.time)
			}
		}
	}()

	gocron.Every(1).Second().Do(task)
	<-gocron.Start()

}
